/*=========================================================================*
 *      File    : bigint.h
 *
 *      Author  : Pallav Gupta
 *      Created : Mon Sep 29 15:25:00 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: bigint.h,v 1.3 2003/12/31 00:47:06 pgupta Exp $ */

#ifndef BIGINT_H
#define BIGINT_H

/* Bit Masks */
#define BIT_ZERO       0x1
#define LOWER_LOWBYTE  0xFF
#define LOWER_HIGHBYTE 0xFF00
#define UPPER_LOWBYTE  0xFF0000
#define UPPER_HIGHBYTE 0xFF000000
#define LOWER_WORD     0xFFFF
#define UPPER_WORD     0xFFFF0000

/* Return codes */
#define BIGINT_ERROR   -1
#define BIGINT_OK      0
#define BIGINT_DIFF    -2 

#define HEXADECIMAL    16
#define BINARY         2

/* BIGINT data structure and macros */
#define HEXBITS        4

#define BYTESIZE       8
#define WORDSIZE       (sizeof(int) * BYTESIZE) /* 32 */
#define NUMBITS        512
#define MAXENTRIES     (NUMBITS/WORDSIZE) /* 16 */
#define MAXWORDINDEX   (MAXENTRIES - 1) /* 15 */
#define MAXBITINDEX    (NUMBITS - 1) /* 511 */

typedef short int INDEX;       /* 16 bits */
typedef unsigned int ELEMENT; /* 32 bits */

typedef struct {
  ELEMENT data[MAXENTRIES]; /* 16 ints, 512 bits */
} BIGINT;


#define CHECK_CONDITION(condition, ret) if (condition)\
                                        return ret

/* External functions interface */
/* Normal routines */
EXTERN BIGINT *bigint_new ARGS((void));
EXTERN BIGINT *bigint_copy ARGS((BIGINT *)); 
EXTERN int bigint_free ARGS((BIGINT *));
EXTERN int bigint_array_free ARGS((BIGINT **, INDEX)); 
EXTERN int bigint_null ARGS((BIGINT *));
EXTERN int bigint_print ARGS((BIGINT *, FILE *));
EXTERN int bigint_compare ARGS((BIGINT *, BIGINT *, INDEX));

/* Conversion routines */
EXTERN BIGINT *ascii_to_bigint ARGS((char *, ELEMENT, ELEMENT));
EXTERN BIGINT **ascii_to_parallel_bigint ARGS((char *, ELEMENT, ELEMENT, ELEMENT));

/* Bit manipulation routines */
EXTERN int bigint_get_bit ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_lower_halfword ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_upper_halfword ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_word ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_lower_lowbyte ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_lower_highbyte ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_upper_lowbyte ARGS((BIGINT *, INDEX));
EXTERN int bigint_get_upper_highbyte ARGS((BIGINT *, INDEX));

EXTERN int bigint_set_bit ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_lower_halfword ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_upper_halfword ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_word ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_lower_lowbyte ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_lower_highbyte ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_upper_lowbyte ARGS((BIGINT *, INDEX, ELEMENT));
EXTERN int bigint_set_upper_highbyte ARGS((BIGINT *, INDEX, ELEMENT));




#endif /* BIGINT_H */
