/*=========================================================================*
 *	File    :	th_int.h
 *
 *	Author  :	Pallav Gupta
 *	Created :	Sun Sep 14 20:44:25 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/


/* $Id: th_int.h,v 1.22 2006/03/30 22:28:17 pgupta Exp $ */

#ifndef TH_INT_H
#define TH_INT_H

#include "bigint.h"
#include <limits.h>
#include <time.h>
#include <math.h>

#define TH_MAGICVALUE 0xabcd1975

#define TH_OTO_FANIN 2 /* one-to-one mapping min fanin */
#define TH_ASSERT(condition) assert(condition)

#define THFACTOR_HICOST 5000000


/* temporary use for majority debug or thsyn debug only */
//#define THRESHOLD_SYN_DEBUG0 1

/* temporary use for state encoding only */
/* STATES_DEBUG1: for area optimization */
/* STATES_DEBUG2: for interconnect optimization */ 
/* STATES_DEBUG3: for node count optimization */
/* each time, just one of the DEBUG flag is on */
//#define STATES_DEBUG1 1
//#define STATES_DEBUG2 1
//#define STATES_DEBUG3 1

/* temporary use for thfactor command only */
//#define THFACTOR_DEBUG0 1
//#define THFACTOR_DEBUG2 1

/* return codes for every command */
enum {
  TH_SUCCESS,
  TH_FAIL,
  TH_ATPG_CONFLICT /* for atpg use only */
};

/* ilp front engine and solver */
#define AMPL "ampl"
#define ILP_SOLVER "cplex"

/* for unate node case in thsyn.c */
#define THSYN_NO_COMBINE 0 
#define THSYN_OR_COMBINE 1
#define THSYN_AND_COMBINE 2

/* size for parallel simulation */
#define TH_SIM32 32
#define TH_MAXWORD 512

#define TH_GETSIMCHANGED(node) (((thnode_t *) node->th)->simchanged)
#define TH_SETSIMCHANGED(node, val) (((thnode_t *) node->th)->simchanged = val)

/* used for simulation */ 
#define TH_SIMSLOT simulation
#define TH_GETVALUE(node) ((int) node->TH_SIMSLOT)
#define TH_SETVALUE(node, val)                 \
  if (TH_GETVALUE(node) != val) {              \
    node->TH_SIMSLOT = (char *) val;           \
    TH_SETSIMCHANGED(node, 1);                 \
  } else                                       \
    TH_SETSIMCHANGED(node, 0)


//#define TH_ATPG_GETLINEVAL(node, index) ((int) ((thnode_t *) node->th)->fanin_lines[index])
//#define TH_ATPG_SETLINEVAL(node, index, value) (((thnode_t *) node->th)->fanin_lines[index] = (int) value)

#define TH_ATPG_GETLINEVALUE(n, i) ((int) ((thnode_t *) n->th)->fanin_lines[i])
#define TH_ATPG_SETLINEVALUE(n, i, v) (((thnode_t *) n->th)->fanin_lines[i] = (int) v)

#define foreach_thfault(faults, gen, f)                 \
   for (gen = lsStart(faults);                          \
        (lsNext(gen, (lsGeneric *) &f, LS_NH) == LS_OK) \
        || ((void) lsFinish(gen), 0); )                        

/* from io/ */
extern int read_line();
extern char *gettoken(char *, char **);

/* thcom.c */
EXTERN void init_threshold ARGS(());
EXTERN void end_threshold ARGS(());
EXTERN void thprint_stats ARGS ((network_t *, FILE *, array_t *));
EXTERN void thprint_buffer ARGS ((network_t *, FILE *, array_t *, int));

/* threshold commands */
EXTERN int com_thbuffer ARGS((network_t **, int, char **));
EXTERN int com_thsyn ARGS((network_t **, int, char **)); 
EXTERN int com_thsyn_with_red_removal ARGS((network_t **, int, char **));
EXTERN int com_majsyn_with_red_removal ARGS((network_t **, int, char **));
EXTERN int com_thconv ARGS((network_t **, int, char **));
EXTERN int com_thsim ARGS((network_t **, int, char **));
EXTERN int com_thprint_buffer ARGS((network_t **, int, char **));
EXTERN int com_thprint ARGS((network_t **, int, char **));
EXTERN int com_thprint_stats ARGS((network_t **, int, char **));
EXTERN int com_thfsim ARGS((network_t **, int, char **));
EXTERN int com_thatpg ARGS((network_t **, int, char **));
EXTERN int com_thvcg ARGS((network_t **, int, char **));
EXTERN int com_thdot ARGS((network_t **, int, char **));
EXTERN int com_thprint_faults ARGS ((network_t **, int argc, char **)); 

/* thatpg_util.c */
EXTERN thatpg_info_t *thatpg_info_init ARGS((network_t *));
EXTERN thatpg_options_t *thatpg_options_init ARGS((int , char **));
EXTERN void thatpg_info_free ARGS((thatpg_info_t *));
EXTERN void thatpg_options_free ARGS((thatpg_options_t *));
EXTERN void thfault_free ARGS((thfault_t *));
EXTERN thfault_t *thfault_init ARGS((node_t *, node_t *, thfault_value_t, thfault_type_t, thatpg_info_t *));
EXTERN void thatpg_add_safaults ARGS((lsList, node_t *, node_t *, thfault_type_t, thatpg_info_t *));
EXTERN int thatpg_gen_faults ARGS ((thatpg_info_t *));
EXTERN void thatpg_print_all_faults ARGS ((FILE *, thatpg_info_t *));
EXTERN void thatpg_print_fault ARGS((FILE *, thfault_t *));
EXTERN void thatpg_print_test_vectors ARGS((thatpg_info_t *));
EXTERN int thatpg_read_test_vectors ARGS((thatpg_info_t *));
EXTERN void thatpg_dump_statistics ARGS((thatpg_info_t *));

/* thfsim.c */
EXTERN array_t *thfsim_simulate_vectors ARGS((thatpg_info_t *)); 
EXTERN void thfsim_simulate_all_faults ARGS((thatpg_info_t *));
EXTERN void thfsim_simulate_fault ARGS((thatpg_info_t *, thfault_t *));
EXTERN void thfsim_inject_fault ARGS((thfault_t *));
EXTERN void thfsim_remove_fault ARGS((thfault_t *));
EXTERN int thfsim_sim_and_compare_vectors ARGS((thatpg_info_t *));


/* thatpg.c */
EXTERN int thatpg_random_test_generation ARGS((thatpg_info_t *));
EXTERN int thatpg_deterministic_test_generation ARGS((thatpg_info_t *));
EXTERN int thatpg_podem ARGS((thatpg_info_t *, thfault_t *));
EXTERN void thatpg_print_values ARGS((network_t *));
EXTERN void thatpg_init_lines ARGS((network_t *));
EXTERN int thatpg_excite_fault ARGS((thfault_t *, array_t *));
EXTERN void thatpg_imply ARGS((thfault_t *, thatpg_assignment_t *, int));
EXTERN void thatpg_simulate_node ARGS((node_t *));
EXTERN thatpg_assignment_t *thatpg_backtrace ARGS((thatpg_assignment_t *));
EXTERN thatpg_assignment_t *thatpg_objective ARGS((thfault_t *));
EXTERN bool thatpg_output_error ARGS((network_t *));
EXTERN bool thatpg_test_possible ARGS((thfault_t *));
EXTERN void thatpg_update_dfrontier ARGS((lsList, node_t *));
EXTERN int thatpg_fault_ilpmodels ARGS((thatpg_info_t *));
EXTERN char *thatpg_ilpcommands ARGS((char *, char *, char *));
EXTERN int thatpg_fault_ilpdata ARGS((thatpg_info_t *, thfault_t *));
EXTERN array_t *thatpg_read_ilpresult ARGS((char *));
EXTERN char *thatpg_convert5 ARGS((thatpg_simv_t));


/* thatpg_cubes.c */
EXTERN int thatpg_calc_pcubes ARGS((network_t *));
EXTERN int thatpg_calc_scovers ARGS((network_t *));

/* thsim.c */
EXTERN thsim_options_t *thsim_options_init ARGS((int, char **));
EXTERN void thsim_options_free ARGS((thsim_options_t *));
EXTERN BIGINT **thsim_comb_thnetwork ARGS((network_t *, BIGINT **, int));

/* thvcg.c */
EXTERN thvcg_options_t *thvcg_options_init ARGS((int, char **));
EXTERN void thvcg_options_free ARGS((thvcg_options_t *));

/* thdot.c */
EXTERN thdot_options_t *thdot_options_init ARGS((int, char **));
EXTERN void thdot_options_free ARGS((thdot_options_t *));

/* thsyn.c */
EXTERN int thsyn_network_syn ARGS((network_t *, thsyn_options_t *));
EXTERN int thsyn_po ARGS ((network_t *, thsyn_options_t *, node_t *));
EXTERN void thsyn_one_node ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN void thsyn_node_collapse ARGS((network_t *, node_t *, int));
EXTERN void thsyn_check_split ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN void thsyn_collapse_w_sharing ARGS((network_t *, node_t *, int));
EXTERN thnode_t *thsyn_create_thnode ARGS((node_t *, thsyn_options_t *, int));
EXTERN void thsyn_unate_node ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN void thsyn_unate_split ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN node_t *thsyn_unate_combine ARGS((node_t *, node_t *, thnode_t *, node_split_t *, thsyn_options_t *));
EXTERN node_t *thsyn_split_w_fanin ARGS((node_t *, thsyn_options_t *));
EXTERN array_t *thsyn_split_binate_node ARGS((node_t *));
EXTERN node_split_t *thsyn_split_unate_node ARGS((node_t *, int));
EXTERN void thsyn_binate_node ARGS((network_t *, thsyn_options_t *, node_t *, int));
EXTERN node_t *thsyn_binate_split_m2 ARGS((node_t *, thsyn_options_t *));

EXTERN int new_thsyn_network_syn ARGS((network_t *, thsyn_options_t *));
EXTERN int new_thsyn_po ARGS ((network_t *, thsyn_options_t *, node_t *));
EXTERN void new_thsyn_one_node ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN void new_thsyn_node_collapse ARGS((network_t *, node_t *, int));
EXTERN void new_thsyn_collapse_w_sharing ARGS((network_t *, node_t *, int));
EXTERN void insert_thnode_to_table ARGS((node_t *));
EXTERN int process_sub_network ARGS((network_t *, node_t *, thsyn_options_t *));
EXTERN array_t *count_subnetwork_andor ARGS((network_t *, node_t *, int));
EXTERN array_t *count_twolevel_node ARGS((network_t *, node_t *,thsyn_options_t *));
EXTERN array_t *new_unate_split ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN node_t *new_thsyn_unate_combine ARGS((node_t *, node_t *, thnode_t *, node_split_t *, thsyn_options_t *));
EXTERN node_t *new_thsyn_split_w_fanin ARGS((node_t *, thsyn_options_t *));
EXTERN array_t *new_binate_split ARGS((network_t *, thsyn_options_t *, node_t *));
EXTERN node_t *new_thsyn_binate_split_m2 ARGS((node_t *, thsyn_options_t *));
EXTERN void thsyn_whole_network ARGS((network_t **, thsyn_options_t *));

/* thsyn_util.c */
EXTERN void thsyn_write_ampl_model ARGS((thsyn_options_t *));
EXTERN void thsyn_write_ampl_commands ARGS((thsyn_options_t *));
EXTERN void thsyn_write_ampl_data ARGS((node_t *, thsyn_options_t *));
EXTERN void thsyn_write_inequalities ARGS((FILE *fp, node_t *, int, int));
EXTERN int thsyn_check_unate ARGS((node_t *));
EXTERN thsyn_options_t *thsyn_options_init ARGS((int, char **));
EXTERN void thsyn_options_free ARGS((thsyn_options_t *));
EXTERN void thsyn_sort_nodes ARGS((array_t *));
EXTERN int thsyn_if_all_PI ARGS((node_t *));
EXTERN double thsyn_area ARGS((network_t *));
EXTERN int thsyn_levels ARGS((network_t *));
EXTERN int thsyn_interconnect ARGS((network_t *));

/* thutil.c */
EXTERN array_t *th_read_test_vectors ARGS((network_t *, FILE *, int *));
EXTERN void print_simresults ARGS((network_t *, FILE *, BIGINT **, int));
EXTERN void array_bigint_free ARGS((array_t *, int)); 
EXTERN int compare_sim_results ARGS((array_t *, array_t *, int, int));

EXTERN char *random_bitvector ARGS((int));
EXTERN unsigned time_seed ARGS(());

/* majsyn.c */
EXTERN int com_majsyn ARGS((network_t **, int, char **));
EXTERN void majsyn_and_or_decomp ARGS((network_t *, node_t *, int, int));
EXTERN void majsyn_or_and_decomp ARGS((network_t *, node_t *, int, int));

EXTERN int majsyn_and_or_try ARGS((node_t *));
EXTERN int majsyn_or_and_try ARGS((node_t *));
EXTERN void majsyn_whole_network ARGS((network_t **, majsyn_options_t *));

/* majsyn_util.c */
EXTERN void majsyn_sort_minterms ARGS((array_t *));
EXTERN void majsyn_uniq_minterms ARGS((array_t *));
EXTERN char *majsyn_create_key ARGS((array_t *));
EXTERN array_t *pfamily_minterms ARGS((pset_family));
EXTERN pset_family pfamily_all_minterms ARGS((pset_family));
EXTERN pset minterm_by_index ARGS((int));
EXTERN int majsyn_array_find_element ARGS((array_t *, int));
EXTERN int two_int_array_same ARGS((array_t *, array_t *));
EXTERN int pattern_is_element ARGS((array_t *, array_t *));
EXTERN array_t *transform_exclude_pfamilys ARGS((array_t *));
EXTERN pset_family comput_off_from_on ARGS((pset_family));
EXTERN pset_family find_must_set ARGS((pset_family, pset_family, pset_family, int));
EXTERN pset_family minterm_indexes_to_pfamily ARGS((array_t *));
EXTERN void majsyn_replace_majnode ARGS((network_t *, node_t *, pset_family *, 
						pset_family *, pset_family *)); 
EXTERN st_table *maj_table_1 ARGS((void)); 
EXTERN st_table *maj_table_2 ARGS((void)); 
EXTERN st_table *maj_table_3 ARGS((void)); 
EXTERN int check_is_majority ARGS((pset_family, st_table *, st_table *, st_table *));
EXTERN majsyn_options_t *majsyn_options_init ARGS((int, char **));
EXTERN void majsyn_options_free ARGS((majsyn_options_t *));

/* majsyn_check.c */
EXTERN int com_majcheck ARGS((network_t **, int, char **));
/* technology mapping for tpl */
EXTERN int com_maptpl ARGS((network_t **, int, char **));

/* my_set_states.c */
EXTERN int com_my_states ARGS((network_t **, int, char **)); 

/* factor based on threshold logic */
EXTERN int com_thfactor ARGS((network_t **, int, char **));
EXTERN void thfactor_split_network ARGS((network_t *, thsyn_options_t *, int));
EXTERN int thfactor_split_node ARGS((network_t *, thsyn_options_t *, node_t *, int));
EXTERN int thfactor_options_init ARGS((int, char **));

/* factor based on majority logic */
EXTERN int com_majfactor ARGS((network_t **, int, char **));
EXTERN void majfactor_split_network ARGS((network_t *, majsyn_options_t *, int));
EXTERN int majfactor_split_node ARGS((network_t *, majsyn_options_t *, node_t *, int));
EXTERN int majfactor_options_init ARGS((int, char **));

/* miscellaneous commands */
EXTERN int com_invcheck ARGS((network_t **, int, char **));


#endif /* TH_INT_H */
