/*=========================================================================*
 *      File    :thsyn.c
 *
 *      Author  : Rui Zhang
 *      Created : Tue Mar 23 17:22:47 2004
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thsyn.c,v 1.12 2006/03/30 22:28:17 pgupta Exp $ */

#include "sis.h"
#include "th_int.h"

static int thsyn_usage ARGS(());
static int thsyn_red_remove_usage ARGS(());
static void fixup_fanin ARGS(());
static void new_fixup_fanin ARGS(());

static st_table *threshold_nodes;
static st_table *new_threshold_nodes;

/************************************************************
 * synthesis main function, calls other functions from here *
 ************************************************************/
int com_thsyn(network_t **network, int argc, char **argv)
{
  thsyn_options_t *opt;  


  if ((opt = thsyn_options_init(argc, argv)) == NIL(thsyn_options_t)
      || opt->help) {
    thsyn_options_free(opt);
    return thsyn_usage();
  }

  thsyn_whole_network(network, opt);
  thsyn_options_free(opt);

  return TH_SUCCESS;
}

/************************************************************
 * synthesis main function with redundancy removal, calls other functions from here *
 ************************************************************/
int com_thsyn_with_red_removal(network_t **network, int argc, char **argv)
{
  thsyn_options_t *opt;

  if ((opt = thsyn_options_init(argc, argv)) == NIL(thsyn_options_t)
       || opt->help) {
    thsyn_options_free(opt);
    return thsyn_red_remove_usage();
  }

  // synthesize the network
  thsyn_whole_network(network, opt);

  // perform redundancy removal
  com_redundancy_removal(network, 1, "red_removal");

  // resynthezie the irredundant network
  thsyn_whole_network(network, opt);

  thsyn_options_free(opt);

  return TH_SUCCESS;
}

void thsyn_whole_network(network_t **network, thsyn_options_t *opt)
{
  network_t *network_oto;
  network_t *threshold_network;
  network_t *new_threshold_network;
  network_t *tmp_network;

  array_t *thnodes;
  array_t *new_thnodes;
  char *key;
  st_generator *gen;
  lsGen lgen;
  node_t *n;
  int i;
  node_t *tmp_node, *po_node;
  int levels;
  double area;
  int connect;

#ifdef THRESHOLD_SYN_DEBUG0 
  fprintf(siserr, "network: %s\n", network_name(*network));
#endif

  /* do one to one mapping */
  network_oto = network_dup(*network);
  th_mapoto(network_oto, opt->max_fanin, opt->max_fanin, opt);
  network_oto->net_name = util_strsav(network_name(*network));
  network_oto->threshold = 1;
  
  /* do synthesis now */
  thsyn_network_syn(*network, opt);
  fixup_fanin();

  thnodes = array_alloc(node_t *, 0);
  st_foreach_item(threshold_nodes, gen, &key, (char **) &n) {
    array_insert_last(node_t *, thnodes, n);
  }

  threshold_network = network_from_nodevec(thnodes);
  threshold_network->net_name = util_strsav(network_name(*network));

  /* adding missing PO nodes */
  /* if a PO appears in a node function, it is not regarded as
     a PO by using network_from_nodevec, we have to add it as
     PO by the following code */
  for (i = 0; i < network_num_po(network_oto); i++) {
    tmp_node = network_get_po(network_oto, i);
    po_node = network_find_node(threshold_network, node_name(tmp_node));
    if (po_node == 0) {
#ifdef THRESHOLD_SYN_DEBUG2 
      printf("can't find node '%s' in threshold node\n", node_name(tmp_node));
#endif
      ;

    }
    else {
      if (node_type(po_node) != PRIMARY_OUTPUT)
	network_add_primary_output(threshold_network, po_node);
    }
  }
 foreach_node(threshold_network, lgen, n) {
   switch (node_function(n)) { 
   case NODE_BUF: 
   case NODE_PI: 
   case NODE_PO: 
     thnode_assign_weights(n,opt);
     break;
   default:
     break;
   }
 }
  
  threshold_network->threshold = 1;

  TH_ASSERT(network_check(threshold_network) && network_is_acyclic(threshold_network));

  network_free(*network);
  *network = network_oto;
  /* do synthesis now */
  /* do threshold synthesis by new method, the starting point is the one-to-one mapped network */
  tmp_network = network_dup(network_oto);
  new_thsyn_network_syn(tmp_network, opt);
  new_fixup_fanin();
  new_thnodes = array_alloc(node_t *, 0);
  st_foreach_item(new_threshold_nodes, gen, &key, (char **) &n) {
    array_insert_last(node_t *, new_thnodes, n);
  }

  new_threshold_network = network_from_nodevec(new_thnodes);
  new_threshold_network->net_name = util_strsav(network_name(*network));
  /* adding missing PO nodes */
  /* if a PO appears in a node function, it is not regarded as
     a PO by using network_from_nodevec, we have to add it as
     PO by the following code */
  for (i = 0; i < network_num_po(network_oto); i++) {
    tmp_node = network_get_po(network_oto, i);
    po_node = network_find_node(new_threshold_network, node_name(tmp_node));
    if (po_node == 0) {
#ifdef THRESHOLD_SYN_DEBUG2 
      printf("can't find node '%s' in threshold node\n", node_name(tmp_node));
#endif
      ;
    }
    else {
      if (node_type(po_node) != PRIMARY_OUTPUT)
	network_add_primary_output(new_threshold_network, po_node);
    }
  }
 foreach_node(new_threshold_network, lgen, n) {
   switch (node_function(n)) { 
   case NODE_BUF: 
   case NODE_PI: 
   case NODE_PO: 
     thnode_assign_weights(n, opt);
     break;
   default:
     break;
   }
 }

  new_threshold_network->threshold = 1;
  TH_ASSERT(network_check(new_threshold_network) && network_is_acyclic(new_threshold_network));

#ifdef THRESHOLD_SYN_DEBUG0
  /* deal with case node:  {a} = pi1, such node is not counted as internal node, should count it */
  int missing_num = 0;
  node_t *check_po, *check_po2;

  /* check every PO node */
  for (i = 0; i < network_num_po(network_oto); i++) {
    check_po = network_get_po(network_oto, i);
    /* error check */
    if (node_num_fanin(check_po) != 1)
    fail("com_thsyn: this node is not a primary output node.");
    /* get the fanin of the primary output */
    check_po2 = node_get_fanin(check_po, 0);
    if (node_type(check_po2) == PRIMARY_INPUT) {
      missing_num++;
      check_po->type = INTERNAL;
    }
  }
  fprintf(siserr, "oto: %d\n", network_num_internal(network_oto) + missing_num );
  area = thsyn_area(network_oto);
  fprintf(siserr, "area1: %.2f\n", area);
  levels = thsyn_levels(network_oto);
  fprintf(siserr, "levels1: %d\n", levels);
  connect = thsyn_interconnect(network_oto);
  fprintf(siserr, "connect1: %d\n", connect);
#endif

#ifdef STATES_DEBUG3
  /* deal with case node:  {a} = pi1, such node is not counted as internal node, should count it */
  int missing_num = 0;
  node_t *check_po, *check_po2;

  /* check every PO node */
  for (i = 0; i < network_num_po(network_oto); i++) {
    check_po = network_get_po(network_oto, i);
    /* error check */
    if (node_num_fanin(check_po) != 1)
    fail("com_thsyn: this node is not a primary output node.");
    /* get the fanin of the primary output */
    check_po2 = node_get_fanin(check_po, 0);
    if (node_type(check_po2) == PRIMARY_INPUT) {
      missing_num++;
      check_po->type = INTERNAL;
    }
  }
#endif 

  if (network_num_internal(threshold_network) < network_num_internal(new_threshold_network)) {
    network_free(*network);
    *network = threshold_network;
    network_free(tmp_network);

#ifdef THRESHOLD_SYN_DEBUG0
    fprintf(siserr, "final: %d\n", network_num_internal(threshold_network));
    area = thsyn_area(threshold_network);
    fprintf(siserr, "area2: %.2f\n", area);
    levels = thsyn_levels(threshold_network);
    fprintf(siserr, "levels2: %d\n", levels);
    connect = thsyn_interconnect(threshold_network);
    fprintf(siserr, "connect2: %d\n", connect);
#endif
  }
  else
    {
      network_free(*network);
      *network = new_threshold_network;
      network_free(tmp_network);

#ifdef THRESHOLD_SYN_DEBUG0
      fprintf(siserr, "final: %d\n", network_num_internal(new_threshold_network));
      area = thsyn_area(new_threshold_network);
      fprintf(siserr, "area2: %.2f\n", area);
      levels = thsyn_levels(new_threshold_network);
      fprintf(siserr, "levels2: %d\n", levels);
      connect = thsyn_interconnect(new_threshold_network);
      fprintf(siserr, "connect2: %d\n", connect);
#endif
    }

#ifdef STATES_DEBUG3
  FILE *fp;
  char *filename;

  filename = ALLOC(char, 100);
  filename = util_strsav("./experiments/ga_thsyn_results.txt");
  if ((fp = fopen(filename, "w")) == NULL) 
    fail("thsyn.c: unable to open file handle for writing."); 
  
  //area optimization
//  area = thsyn_area(*network);  
//  fprintf(fp, "%.2f\n", area);  
  
  //interconnect optimization
//  connect = thsyn_interconnect(*network); 
//  fprintf(fp, "%d\n", connect);   
  
  //node count optimization
  fprintf(fp, "%d\n", network_num_internal(*network) + missing_num );   
    
  fclose(fp);
  if (filename != NIL(char))
    FREE(filename);
#endif

#ifdef STATES_DEBUG1
  FILE *fp;
  char *filename;

  filename = ALLOC(char, 100);
  filename = util_strsav("./experiments/ga_thsyn_results.txt");
  if ((fp = fopen(filename, "w")) == NULL) 
    fail("thsyn.c: unable to open file handle for writing."); 
  
  //area optimization
  area = thsyn_area(*network);  
  fprintf(fp, "%.2f\n", area);  
  fclose(fp);
  if (filename != NIL(char))
    FREE(filename);
#endif


#ifdef STATES_DEBUG2
  FILE *fp;
  char *filename;

  filename = ALLOC(char, 100);
  filename = util_strsav("./experiments/ga_thsyn_results.txt");
  if ((fp = fopen(filename, "w")) == NULL) 
    fail("thsyn.c: unable to open file handle for writing."); 
  
  //interconnect optimization
  connect = thsyn_interconnect(*network); 
  fprintf(fp, "%d\n", connect);   
  fclose(fp);
  if (filename != NIL(char))
    FREE(filename);
#endif


}

/********************************************************/
/* the main body of our threshold synthesis methodology */
/********************************************************/
int new_thsyn_network_syn(network_t *network, thsyn_options_t *opt)
{
  int i;
  node_t *n;

  /* write the linear problem AMPL model file and command files */
  thsyn_write_ampl_model(opt);
  thsyn_write_ampl_commands(opt);

  /* create the hash table of threshold nodes */
  new_threshold_nodes = st_init_table(strcmp, st_strhash);

  /* synthesize one output at a time */
  for (i = 0; i < network_num_po(network); i++) {
    n = network_get_po(network, i);
    TH_ASSERT(new_thsyn_po(network, opt, n) == TH_SUCCESS);
  }
  
  return TH_SUCCESS;
}

/*************************************
 * synthesis one primary output node *
 *************************************/
int new_thsyn_po(network_t *network, thsyn_options_t *opt, node_t *n)
{
  node_t *ni;

  /* error check */
  if (node_num_fanin(n) != 1)
    fail("thsyn_po: this node is not a primary output node.");
  
  /* get the fanin of the primary output */
  ni = node_get_fanin(n, 0);

  if (node_type(ni) != PRIMARY_INPUT)
    new_thsyn_one_node(network, opt, ni);
  else {
    /* deal with special case when a PO node is also a PI node */
    thnode_t *th;
    node_t *new_node;
    
    new_node = node_and(node_constant(1), node_literal(ni,1));
    FREE(new_node->name);
    new_node->name = util_strsav(node_name(n));
    
    /* write AMPL data file  */
    thsyn_write_ampl_data(new_node, opt);
    
    /* system call to AMPL */
    if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
      fail("new_thsyn_po: system call failed.");
    
    th = thsyn_create_thnode(new_node, opt, THSYN_NO_COMBINE); 
    /* attach the thnode to the node */
    new_node->th = (char *) th;
    insert_thnode_to_table(new_node);
  }
  return TH_SUCCESS;
}

/**********************
 * synthesis one node *
 **********************/
void new_thsyn_one_node(network_t *network, thsyn_options_t *opt, node_t *n)
{
  int i;
  int stat;
  node_t *fanin;


  if (node_type(n) == PRIMARY_INPUT) {
    return;
  }
  /* no need to thsynthesis if this node is already a threshold 
     node in the synthesized threshold network */
  if (st_lookup(new_threshold_nodes, node_name(n), NIL(char *))) 
    return;
  stat = process_sub_network(network, n, opt);
  if (stat == 0)
    return;

  foreach_fanin(n, i, fanin) {
    new_thsyn_one_node(network, opt, fanin);
  }
}

/***************************************************************************/
/* returns 0: when this subnetwork has been processed before, 1: otherwise */
/***************************************************************************/
int process_sub_network(network_t *network, node_t *n, thsyn_options_t *opt)
{
  node_t *tmp_node, *node_copy, *t_node;
  int num_andor; /* num of and_or gates in original subnetwork */
  int num_split; /* num of split threshold nodes needed by our splitting method */
  int j;
  array_t *and_or_array;
  array_t *split_nodes_array;

  /* the subnetwork is got by collapsing the node based on maximum fanin */

  /* first count the number of and-or gates in the subnetwork */
  /* save the and_or gates information in and_or_array */

  tmp_node = node_dup(n);
  FREE(tmp_node->name);
  tmp_node->name = util_strsav(node_name(n));
  and_or_array = count_subnetwork_andor(network, tmp_node, opt->max_fanin);

  num_andor = array_n(and_or_array);
  if (num_andor == 0)
    return 0;

  /* next, get the nodes information needed by our splitting method */
  /* save the nodes information in split nodes array */

  new_thsyn_node_collapse(network, n, opt->max_fanin);
  node_copy = node_dup(n);
  FREE(node_copy->name);
  node_copy->name = util_strsav(node_name(n));
  split_nodes_array = count_twolevel_node(network, node_copy, opt);
  num_split = array_n(split_nodes_array);
  if (num_split <=0)
    fail("process_sub_network: after splitting, got no split nodes");
  /* compare two methods, pick the smaller one */
  if (num_andor <= num_split) {
    for (j = 0; j < num_andor; j++) {
      t_node = array_fetch(node_t *,and_or_array,j);
      insert_thnode_to_table(t_node);
    }
  }
  else {
    for (j = 0; j < num_split; j++) {
      t_node = array_fetch(node_t *,split_nodes_array,j);
      insert_thnode_to_table(t_node);
    }
  }
  return 1;
}

array_t *count_twolevel_node(network_t *network, node_t *n, thsyn_options_t *opt)
{ 
  array_t *split_nodes;
  int is_unate;

  split_nodes = array_alloc(node_t *, 0);
  /* check unateness */
  is_unate = thsyn_check_unate(n);
      if (is_unate == BINATE) 
	split_nodes = new_binate_split(network, opt, n);
      else
	split_nodes = new_unate_split(network, opt, n);

  return split_nodes;
}


array_t *new_binate_split(network_t *network, thsyn_options_t *opt, node_t *n)
{
  int i, unate;
  array_t *array_large, *array_small;
  node_t *combined_node, *tmp_node, *fanin;
  thnode_t *th;

  array_large = array_alloc(node_t *, 0);
  array_small = array_alloc(node_t *,0);
  combined_node = new_thsyn_binate_split_m2(n, opt);
  /* save this threshold node in array_large */
  array_insert_last(node_t *, array_large, combined_node);
  foreach_fanin(combined_node, i, fanin) {
	if(node_num_cube(fanin) == 1) {
	  /* write AMPL data file  */
	  thsyn_write_ampl_data(fanin, opt);
	  /* system call to AMPL */
	  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	    fail("new_binate_split: system call failed.");
	  th = thsyn_create_thnode(fanin, opt, THSYN_NO_COMBINE); 
	  if (th == NIL(thnode_t)) {
	    fail("new_binate_split: AMPL failed to solve a AND gate");
	  }
	  else {
	    /* create a threshold node based on thnode, and add this thnode into threshold network */
	    tmp_node = node_dup(fanin);
	    /* attach the thnode to the node */
	    tmp_node->th = (char *) th;
	    FREE(tmp_node->name);
	    tmp_node->name = util_strsav(node_name(fanin));
	    /* save this threshold node in big_array */
	    array_insert_last(node_t *, array_large, tmp_node);
	  }
	}
	else { /* fanin is not simply a AND gate, continuing checking and splitting */
	  unate = thsyn_check_unate(fanin);
	  if (unate == BINATE) 
	    array_small = new_binate_split(network, opt, fanin); 
	  else 
	    array_small = new_unate_split(network, opt, fanin);
	  if (array_n(array_small) <= 0)
	    fail("no contents in array_small");
	  array_append(array_large, array_small);
	}
  }
  return array_large;
}

array_t *new_unate_split(network_t *network, thsyn_options_t *opt, node_t *n)
{ 
  array_t *big_array, *small_array;
  thnode_t *th;
  node_t *tmp_node;
  node_split_t *split;
  int case_index, unate, i;
  node_t *sn1, *sn2, *combined_node, *sn1_literal, *fanin;
  

  big_array = array_alloc(node_t *, 0);
  small_array = array_alloc(node_t *,0);
  /* write AMPL data file  and call AMPL*/
  thsyn_write_ampl_data(n, opt);
  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
    fail("thsyn_unate_node: system call failed");
  /* parse AMPL result: call this function to save the returned weight-threshold 
     value to a threshold node, and save this threshold node to a network */
  th = thsyn_create_thnode(n, opt, THSYN_NO_COMBINE); 
  if (th == NIL(thnode_t)) { /* this unate node is NOT a threshold node */
    /* first split into half to check */
    split = thsyn_split_unate_node(n, 1 /*allow and-split case*/);
    /* set the index: whether it was an OR split or AND split */
    if (split->split_flag)
      case_index = THSYN_AND_COMBINE;
    else
      case_index = THSYN_OR_COMBINE;
    
    /* sort the nodes and get the splitted nodes in descending order */
    thsyn_sort_nodes(split->nodes);
    sn1 = array_fetch(node_t *, split->nodes, 0);
    sn2 = array_fetch(node_t *, split->nodes, 1);
    
    /* check if sn1 or sn2 is unate, fail if we get BINATE */
    if ((thsyn_check_unate(sn1) == BINATE) || (thsyn_check_unate(sn2) == BINATE))
      fail("new_unate_split: encountered binate node in unate split");
    
    /* check if sn2 fans out to only one primary input, if it is, for example, 
       f = sn1 sn2, where sn2 fans out to only one primary input: sn2 = pa,
       store f: sn1, pa, [2 1 1], then process fanins of sn1 recursively. 
       This can't happen for sn1, since sn1 and sn2 are sorted in descending order,
       If sn1 only has one fanin, sn2 can only has one fanin, and all two fanin nodes
       are threshold node. */
    if (node_num_fanin(sn2) == 1) {
      sn1_literal = node_literal(sn1, 1); /* 0: complemented, 1: uncomplemented */
      if (case_index == THSYN_OR_COMBINE)  /* OR case */
	combined_node = node_or(sn1_literal, sn2);
      else /* AND case */
	combined_node = node_and(sn1_literal, sn2);
      
      FREE(combined_node->name);
      combined_node->name = util_strsav(split->pn_name);
      
      /* write AMPL data file  */
      thsyn_write_ampl_data(combined_node, opt);
      
      /* system call to AMPL */
      if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	fail("new_unate_split: system call failed.");
      
      th = thsyn_create_thnode(combined_node, opt, THSYN_NO_COMBINE); 
      
      if (th == NIL(thnode_t)) {
	fail("new_unate_split: AMPL failed to solve a 2-input OR gate");
      }
      else {
	/* create a threshold node based on thnode, and add this thnode into threshold network */
	tmp_node = node_dup(combined_node);
	/* attach the thnode to the node */
	tmp_node->th = (char *) th;
	FREE(tmp_node->name);
	tmp_node->name = util_strsav(node_name(combined_node));
	/* save this threshold node in big_array */
	array_insert_last(node_t *, big_array, tmp_node);
	unate = thsyn_check_unate(sn1);
	if (unate == BINATE)
	  fail("meet binate node after splitting unate node");
	/* then, recursive call function new_unate_split */
	small_array = new_unate_split(network, opt, sn1);
	if (array_n(small_array) <= 0)
	  fail("no contents in small_array");
	array_append(big_array, small_array);
	
      }
    }
    else {
      /* for sn1 and sn2 are nodes that have more than one fanin, call AMPL on the larger node,
	 check if it is a threshold node, if not, call AMPL on the smaller node, and check again */
      if (case_index != THSYN_OR_COMBINE)
	fail("meet THSYN_AND_COMBINE case, which is impossible here");
      
      /* write AMPL data file for sn1 */
      thsyn_write_ampl_data(sn1, opt);
      /* system call to AMPL */
      if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	fail("new_unate_split: system call failed.");
      
      th = thsyn_create_thnode(sn1, opt, THSYN_OR_COMBINE); /* it is always THSYN_OR_COMBINE case */
      if (th != NIL(thnode_t)) {
	/* sn1 is a threshold node, then get new threshold pn_name = sn1(...) + sn2 */
	combined_node = new_thsyn_unate_combine(sn1, sn2, th, split, opt);
	tmp_node = node_dup(combined_node);
	FREE(tmp_node->name);
	tmp_node->name = util_strsav(node_name(combined_node));
	array_insert_last(node_t *, big_array, tmp_node);
	/* process sn2, if sn2 is simply a AND gate, add this threshold node into big_array */
	if(node_num_cube(sn2) == 1) {
	  /* write AMPL data file  */
	  thsyn_write_ampl_data(sn2, opt);
	  /* system call to AMPL */
	  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	    fail("new_unate_split: system call failed.");
	  th = thsyn_create_thnode(sn2, opt, THSYN_NO_COMBINE); 
	  if (th == NIL(thnode_t)) {
	    fail("new_unate_split: AMPL failed to solve a AND gate");
	  }
	  else {
	    /* create a threshold node based on thnode, and add this thnode into threshold network */
	    tmp_node = node_dup(sn2);
	    /* attach the thnode to the node */
	    tmp_node->th = (char *) th;
	    FREE(tmp_node->name);
	    tmp_node->name = util_strsav(node_name(sn2));
	    /* save this threshold node in big_array */
	    array_insert_last(node_t *, big_array, tmp_node);
	  }
	}
	else { /* sn2 is not simply a AND gate, and since */
	  /* sn2 is not a threshold node, needs continuing splitting unate node sn2 */
	  /* first, make sure sn2 is a unate node */
	  unate = thsyn_check_unate(sn2);
	  if (unate == BINATE)
	    fail("meet binate node after splitting unate node");
	  /* then, recursive call function new_unate_split */
	  small_array = new_unate_split(network, opt, sn2);
	  if (array_n(small_array) <= 0)
	    fail("no contents in small_array");
	  array_append(big_array, small_array);
	}
      }
      else { /* sn1 is not a threshold node, go to check sn2 */
	/* write AMPL data file for sn2 */
	thsyn_write_ampl_data(sn2, opt);
	/* system call to AMPL */
	if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	  fail("new_unate_split: system call failed.");
	
	th = thsyn_create_thnode(sn2, opt, THSYN_OR_COMBINE); /* it is always THSYN_OR_COMBINE case */
	if (th != NIL(thnode_t)) {
	  /* sn2 is a threshold node, then get new threshold pn_name = sn1 + sn2(...) */
	  combined_node = new_thsyn_unate_combine(sn2, sn1, th, split, opt);
	  tmp_node = node_dup(combined_node);
	  FREE(tmp_node->name);
	  tmp_node->name = util_strsav(node_name(combined_node));
	  array_insert_last(node_t *, big_array, tmp_node);
	  /* process sn1, if sn1 is simply a AND gate, add this threshold node into big_array */
	  if(node_num_cube(sn1) == 1) {
	    /* write AMPL data file  */
	    thsyn_write_ampl_data(sn1, opt);
	    /* system call to AMPL */
	    if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	      fail("new_unate_split: system call failed.");
	    th = thsyn_create_thnode(sn1, opt, THSYN_NO_COMBINE); 
	    if (th == NIL(thnode_t)) {
	      fail("new_unate_split: AMPL failed to solve a AND gate");
	    }
	    else {
	      /* create a threshold node based on thnode, and add this thnode into threshold network */
	      tmp_node = node_dup(sn1);
	      /* attach the thnode to the node */
	      tmp_node->th = (char *) th;
	      FREE(tmp_node->name);
	      tmp_node->name = util_strsav(node_name(sn1));
	      /* save this threshold node in big_array */
	      array_insert_last(node_t *, big_array, tmp_node);
	    }
	  }
	  else { /* sn1 is not simply a AND gate, and since sn1 is not a threshold node, */
	         /* needs continuing splitting unate node sn1 */
	         /* first, make sure sn1 is a unate node */
	    unate = thsyn_check_unate(sn1);
	    if (unate == BINATE)
	      fail("meet binate node after splitting unate node");
	    /* then, recursive call function new_unate_split */
	    small_array = new_unate_split(network, opt, sn1);
	    if (array_n(small_array) <= 0)
	      fail("no contents in small_array");
	    array_append(big_array, small_array);
	  }
	}
	/* if half doesn't work, split into num_cubes nodes, also add each cube into the array */
	else { /* both sn1 and sn2 are not threshold case */
	  combined_node = new_thsyn_split_w_fanin(n, opt);
	  tmp_node = node_dup(combined_node);
	  FREE(tmp_node->name);
	  tmp_node->name = util_strsav(node_name(combined_node));
	  /* check */
	  if (node_num_cube(n) !=  node_num_cube(tmp_node))
	    fail("didn't split according to the num of cubes.");
	  array_insert_last(node_t *, big_array, tmp_node);
	  /* add every cube(threshold node already) into the big_array */
	  foreach_fanin(combined_node, i, fanin) {
	    /* write AMPL data file  */
	    thsyn_write_ampl_data(fanin, opt);
	    /* system call to AMPL */
	    if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	      fail("new_unate_split: system call failed.");
	    th = thsyn_create_thnode(fanin, opt, THSYN_NO_COMBINE); 
	    if (th == NIL(thnode_t))
	      fail("ampl returned no result for AND gate");
	    /* attach the thnode to the node */
	    tmp_node = node_dup(fanin);
	    FREE(tmp_node->name);
	    tmp_node->name = util_strsav(node_name(fanin));
	    tmp_node->th = (char *) th;
	    array_insert_last(node_t *, big_array, tmp_node);  
	  }
	}
      }
    }
  }
  else {
    /* create a threshold node based on thnode, and add this thnode into big_array */
    tmp_node = node_dup(n);
    tmp_node->th = (char *) th;
    FREE(tmp_node->name);
    tmp_node->name = util_strsav(node_name(n));
    array_insert_last(node_t *, big_array, tmp_node);
  } 
  
  return big_array;
}


array_t *count_subnetwork_andor(network_t *network, node_t *n, int max)
{
  array_t *and_or;
  int i, nfanin, flag;
  node_t *copy, *copy1, *fanin, *node_in_network;
  node_t *p;

  nfanin = node_num_fanin(n);
  and_or = array_alloc(node_t *, 0);
  if (nfanin == 0 || nfanin > max) {
    printf("original failed, nfanin = %d, max = %d\n", nfanin, max);
    //rui    fail("count_subnework_andor: original node exceeds max fanin restriction or = 0.");
  }

  /* check if this node exists in the current threshold network or not */
  if (st_lookup(new_threshold_nodes, node_name(n), NIL(char *))) {
    return and_or;
  }
  /* no collapsing to be done if fanin is 0 or greater than limit */
  if (nfanin == max) {
    array_insert_last(node_t *, and_or, n);
    return and_or;
  }

  /* save the root and_or node in and_or array (it is a threshold node already) */
  p = node_dup(n);
  FREE(p->name);
  p->name = util_strsav(node_name(n));
  array_insert_last(node_t *, and_or, p);

  /* while fanin is less than limit and all fanin are NOT primary inputs */
  while (nfanin < max && !thsyn_if_all_PI(n)) {
    copy = node_dup(n);
    flag = 0;
    foreach_fanin(copy, i, fanin) {
      /* check if this fanin exists in the current threshold network or not */
      /* if it is already a threshold node in threshold network, do not collapse on it */
      if (!st_lookup(new_threshold_nodes, node_name(fanin), NIL(char *))) {  
	if (node_type(fanin) != PRIMARY_INPUT) {
	  /* check if this fanin is a node in the original network */
	  /* if the fanin node is not in the original network (newly created node),
	     do not call function node_num_fanout, otherwise, problem */
	  node_in_network = network_find_node(network, node_name(fanin));
	  if (node_in_network != 0) {
	    if (node_num_fanout(node_in_network) == 1) { 
	      /* this fanin only fanout to one other node, no sharing */
	      copy1 = node_dup(n);
	      node_collapse(copy1, fanin);
	      if (node_num_fanin(copy1) <= max &&
		  !node_equal(copy1, node_constant(0)) &&
		  !node_equal(copy1, node_constant(1))) {
		node_collapse(n, fanin);
		flag = 1;
		// save the original and_or node in and_or array 
		p = node_dup(fanin);
		FREE(p->name);
		p->name = util_strsav(node_name(fanin));
		array_insert_last(node_t *, and_or, p);
	      }       	
	      node_free(copy1);
	    }
	  }
	  else { 
	    /* just collapse on this fanin without calling node_num_fanout */
	    copy1 = node_dup(n);
	    node_collapse(copy1, fanin);
	    if (node_num_fanin(copy1) <= max &&
		!node_equal(copy1, node_constant(0)) &&
		!node_equal(copy1, node_constant(1))) {
	      node_collapse(n, fanin);
	      flag = 1;
	      /* save the original and_or node in and_or array */
	      p = node_dup(fanin);
	      FREE(p->name);
	      p->name = util_strsav(node_name(fanin));
	      array_insert_last(node_t *, and_or, p);
	    }       	
	    node_free(copy1);
	  }
	}
      }
      nfanin = node_num_fanin(n);
      
      if (nfanin == 0) {
	fail("num fanin is zero");
      }      
      if (nfanin > max)
	break;
    }
    node_free(copy);
    if (!flag)
      break;
  }

  return and_or;
}

/******************************************************** 
 * collapse the node based on the max fanin restriction *
 ********************************************************/
void new_thsyn_node_collapse(network_t *network, node_t *n, int max)
{
  int num_fanin;

  /* if the current fanins are less than max fanin restriction, continue collapsing */
  num_fanin = node_num_fanin(n);

  if (num_fanin <= max) 
    new_thsyn_collapse_w_sharing(network, n, max);
  /* if the current fanins are more than max fanin restriction, split */
  else 
    ; 
}

void new_thsyn_collapse_w_sharing(network_t *network, node_t *n, int max)
{
  int i, nfanin, flag;
  node_t *copy, *copy1, *fanin, *node_in_network;

  nfanin = node_num_fanin(n);
  /* no collapsing to be done if this node is already a threshold 
     node in the synthesized threshold network */
  /* check if this node exists in the current threshold network or not */
  if (st_lookup(new_threshold_nodes, node_name(n), NIL(char *))) {
    return;
  }
  /* no collapsing to be done if fanin is 0 or greater than limit */
  if (nfanin == 0 || nfanin >= max) {
    return;
  }
  /* while fanin is less than limit and all fanin are NOT primary inputs */
  while (nfanin < max && !thsyn_if_all_PI(n)) {
    copy = node_dup(n);
    flag = 0;
    foreach_fanin(copy, i, fanin) {
      /* check if this fanin exists in the current threshold network or not */
      /* if it is already a threshold node in threshold network, do not collapse on it */
      if (!st_lookup(new_threshold_nodes, node_name(fanin), NIL(char *))) {  
	if (node_type(fanin) != PRIMARY_INPUT) {
	  /* check if this fanin is a node in the original network */
	  /* if the fanin node is not in the original network (newly created node),
	     do not call function node_num_fanout, otherwise, problem */
	  node_in_network = network_find_node(network, node_name(fanin));
	  if (node_in_network != 0) {
	    if (node_num_fanout(node_in_network) == 1) { 
	      /* this fanin only fanout to one other node, no sharing */
	      copy1 = node_dup(n);
	      node_collapse(copy1, fanin);
	      if (node_num_fanin(copy1) <= max &&
		  !node_equal(copy1, node_constant(0)) &&
		  !node_equal(copy1, node_constant(1))) {
		node_collapse(n, fanin);
		flag = 1;
	      }       	
	      node_free(copy1);
	    }
	  }
	  else { 
	    /* just collapse on this fanin without calling node_num_fanout */
	    copy1 = node_dup(n);
	    node_collapse(copy1, fanin);
	    if (node_num_fanin(copy1) <= max &&
		!node_equal(copy1, node_constant(0)) &&
		!node_equal(copy1, node_constant(1))) {
	      node_collapse(n, fanin);
	      flag = 1;
	    }       	
	    node_free(copy1);
	  }
	}
      }
      nfanin = node_num_fanin(n);
      
      if (nfanin == 0) {
	fail("num fanin is zero");
      }      
      if (nfanin > max)
	break;
    }
    node_free(copy);
    if (!flag)
      break;
  }
}

void insert_thnode_to_table(node_t *n)
{
  node_t *tmp_node;

  tmp_node = node_dup(n);
  FREE(tmp_node->name);
  tmp_node->name = util_strsav(node_name(n));
  tmp_node->type = INTERNAL;

  /* check if this node exists in the current threshold network or not */
  if (!st_lookup(new_threshold_nodes, node_name(tmp_node), NIL(char *))) {
    /* save this threshold node in threshold network */
    st_insert(new_threshold_nodes, node_name(tmp_node), (char *) tmp_node);
  }
}

/* the main body of our threshold synthesis methodology */
int thsyn_network_syn(network_t *network, thsyn_options_t *opt)
{
  int i;
  node_t *n;
  array_t *node_vec;

  /* preprocessing network to satisfy max_fanin restriction */
  node_vec = network_dfs(network);
  for (i = 0; i < array_n(node_vec); i++) {
    n = array_fetch(node_t *,node_vec,i);
    if (node_num_fanin(n) > opt->max_fanin)
      majsyn_and_or_decomp(network, n, /* and_limit */ opt->max_fanin, /* or_limit */ opt->max_fanin); 
  }
  
  /* write the linear problem AMPL model file and command files */
  thsyn_write_ampl_model(opt);
  thsyn_write_ampl_commands(opt);

  /* create the hash table of threshold nodes */
  threshold_nodes = st_init_table(strcmp, st_strhash);

  /* synthesize one output at a time */
  for (i = 0; i < network_num_po(network); i++) {
    n = network_get_po(network, i);
    TH_ASSERT(thsyn_po(network, opt, n) == TH_SUCCESS);
  }
  
  return TH_SUCCESS;
}

/*************************************
 * synthesis one primary output node *
 *************************************/
int thsyn_po(network_t *network, thsyn_options_t *opt, node_t *n)
{
  node_t *ni;

  /* error check */
  if (node_num_fanin(n) != 1)
    fail("thsyn_po: this node is not a primary output node.");
  
  /* get the fanin of the primary output */
  ni = node_get_fanin(n, 0);

  if (node_type(ni) != PRIMARY_INPUT)
    thsyn_one_node(network, opt, ni);
  else {
    /* deal with special case when a PO node is also a PI node */

    thnode_t *th;
    node_t *new_node;

    new_node = node_and(node_constant(1), node_literal(ni,1));
    FREE(new_node->name);
    new_node->name = util_strsav(node_name(n));

    /* write AMPL data file  */
    thsyn_write_ampl_data(new_node, opt);
    
    /* system call to AMPL */
    if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
      fail("thsyn_unate_split: system call failed.");
    
    th = thsyn_create_thnode(new_node, opt, THSYN_NO_COMBINE); 
    /* attach the thnode to the node */
    new_node->th = (char *) th;
    if (!st_lookup(threshold_nodes, node_name(new_node), NIL(char *))) {
      /* save this threshold node in threshold network */
      st_insert(threshold_nodes, node_name(new_node), (char *) new_node);
    }
  }
  
    return TH_SUCCESS;
}

/**********************
 * synthesis one node *
 **********************/
void thsyn_one_node(network_t *network, thsyn_options_t *opt, node_t *n)
{
  if (node_type(n) == PRIMARY_INPUT)
    return;

  /* no need to thsynthesis if this node is already a threshold 
     node in the synthesized threshold network */
  if (st_lookup(threshold_nodes, node_name(n), NIL(char *))) 
    return;
  
  /* collapse the node based on the max fanin restriction */
    thsyn_node_collapse(network, n, opt->max_fanin);
    thsyn_check_split(network, opt, n);
}

/******************************************************** 
 * collapse the node based on the max fanin restriction *
 ********************************************************/
void thsyn_node_collapse(network_t *network, node_t *n, int max)
{
  int num_fanin;

  /* if the current fanins are less than max fanin restriction, continue collapsing */
  num_fanin = node_num_fanin(n);

  if (num_fanin <= max) 
    thsyn_collapse_w_sharing(network, n, max);
  /* if the current fanins are more than max fanin restriction, split */
  else 
    ; 
}

/********************************************************************* 
 * this function synthesize a given node, check if it is a threshold *
 * function, if not, split it to create threshold nodes if possible  *
 *********************************************************************/
void thsyn_check_split(network_t *network, thsyn_options_t *opt, node_t *n)
{
  int unate;

  /* no need to synthesis if this node is already a threshold 
     node in the synthesized threshold network */
  /* check if this node exists in the current threshold network or not */
  if (st_lookup(threshold_nodes, node_name(n), NIL(char *)))
    return; 
  
  unate = thsyn_check_unate(n);
  if (unate == BINATE)
    thsyn_binate_node(network, opt, n, 2); 
  else 
    thsyn_unate_node(network, opt, n);
}

/***************************************************************************************************************
 * This function collapses the node. Given a node and max fanin restriction 'm' as parameters to this function,*
 * it will collapse the node one level back, and check if the number of fanins is greater than 'm' and also    *
 * check if all the fanins are primary inputs. If one of the condition meets, stop collapsing, otherwise,      *
 * continue collapsing one level back again, and check the same conditions and so on.                          *
 ***************************************************************************************************************/
void thsyn_collapse_w_sharing(network_t *network, node_t *n, int max)
{
  int i, nfanin, flag;
  node_t *copy, *copy1, *fanin, *node_in_network;

  nfanin = node_num_fanin(n);
  /* no collapsing to be done if this node is already a threshold 
     node in the synthesized threshold network */
  /* check if this node exists in the current threshold network or not */
    if (st_lookup(threshold_nodes, node_name(n), NIL(char *))) 
      return;

  /* no collapsing to be done if fanin is 0 or greater than limit */
  if (nfanin == 0 || nfanin >= max) 
    return;
  
  /* while fanin is less than limit and all fanin are NOT primary inputs */
  while (nfanin < max && !thsyn_if_all_PI(n)) {
    copy = node_dup(n);
    flag = 0;
    foreach_fanin(copy, i, fanin) {
      /* check if this fanin exists in the current threshold network or not */
      /* if it is already a threshold node in threshold network, do not collapse on it */
      if (!st_lookup(threshold_nodes, node_name(fanin), NIL(char *))) {  
	if (node_type(fanin) != PRIMARY_INPUT) {
	  /* check if this fanin is a node in the original network */
	  /* if the fanin node is not in the original network (newly created node),
	     do not call function node_num_fanout, otherwise, problem */
	  node_in_network = network_find_node(network, node_name(fanin));
	  if (node_in_network != 0) {
	    if (node_num_fanout(node_in_network) == 1) { 
	      /* this fanin only fanout to one other node, no sharing */
	      copy1 = node_dup(n);
	      node_collapse(copy1, fanin);
	      if (node_num_fanin(copy1) <= max &&
		  !node_equal(copy1, node_constant(0)) &&
		  !node_equal(copy1, node_constant(1))) {
		node_collapse(n, fanin);
		flag = 1;
	      }       	
	      node_free(copy1);
	    }
	  }
	  else { 
	    /* just collapse on this fanin without calling node_num_fanout */
	    copy1 = node_dup(n);
	    node_collapse(copy1, fanin);
	    if (node_num_fanin(copy1) <= max &&
		!node_equal(copy1, node_constant(0)) &&
		!node_equal(copy1, node_constant(1))) {
	      node_collapse(n, fanin);
	      flag = 1;
	    }       	
	    node_free(copy1);
	  }
	}
      }
      nfanin = node_num_fanin(n);
      
      if (nfanin == 0) {
	fail("num fanin is zero");
      }      
      if (nfanin > max)
	break;
    }
    node_free(copy);
    if (!flag)
      break;
  }
}

/**************************************************
 * this function thsynthesize a given unate node  *
 **************************************************/
void thsyn_unate_node(network_t *network, thsyn_options_t *opt, node_t *n)
{
  int num_fanin, i;
  thnode_t *th;
  node_t *tmp_node;

  if (node_type(n) != PRIMARY_INPUT) {
    /* write AMPL data file  and call AMPL*/
    thsyn_write_ampl_data(n, opt);
    if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
      fail("thsyn_unate_node: system call failed");
    /* parse AMPL result: call this function to save the returned weight-threshold 
       value to a threshold node, and save this threshold node to a network */
    th = thsyn_create_thnode(n, opt, THSYN_NO_COMBINE); 
    if (th == NIL(thnode_t)) { /* this unate node is NOT a threshold node */
      thsyn_unate_split(network, opt, n);
    }
    else {
      /* create a threshold node based on thnode, and add this thnode into threshold network */
      tmp_node = node_dup(n);
      tmp_node->th = (char *) th;
      FREE(tmp_node->name);
      tmp_node->name = util_strsav(node_name(n));

      /* check if this node exists in the current threshold network or not */
      if (!st_lookup(threshold_nodes, node_name(tmp_node), NIL(char *))) {
	/* save this threshold node in threshold network */
	st_insert(threshold_nodes, node_name(tmp_node), (char *) tmp_node);
      }
      
      /* for each fanin of tmp_node, start thsyn_one_node again */
      num_fanin = node_num_fanin(n);
      for ( i = 0; i < num_fanin; i++)
	thsyn_one_node(network, opt, node_get_fanin(n, i));
    } 
  }
}

/**************************************************
 * this function thsynthesize a given binate node *
 **************************************************/
void thsyn_binate_node(network_t *network, thsyn_options_t *opt, node_t *node, int binate_method)
{
  int num_fanin, i, is_unate;
  node_t *combined_node, *fanin;
  

  if (binate_method == 1) { 
    /* split binate node according to method 1 */
    /* i.e., simply split it into m-input OR gate */
    combined_node = thsyn_split_w_fanin(node, opt);
    
    /* for each fanin of tmp_node, start thsyn_one_node again */
    num_fanin = node_num_fanin(combined_node);
    for (i = 0; i < num_fanin; i++)
      thsyn_one_node(network, opt, node_get_fanin(combined_node,i));
  }
  else if (binate_method == 2) {
    /* split binate node according to method 2, i.e., split it into 
       at most m-input OR gate, stop when there's a unate node */
    
    combined_node = thsyn_binate_split_m2(node, opt);
    foreach_fanin(combined_node, i, fanin) {
      is_unate = thsyn_check_unate(fanin);
      if (is_unate == BINATE) 
	thsyn_one_node(network, opt, fanin);
      else
	thsyn_unate_node(network, opt, fanin);
    }
  }
}

/******************************************************************************
 * given a node, this function calls AMPl to check if it is a threshold node, *
 * if no, returns NULL, otherwise returns thnode_t *th, which contains the    *
 * weights and threshold values of the node                                   *
 ******************************************************************************/
thnode_t *thsyn_create_thnode(node_t *n, thsyn_options_t *opt, int combine_case)
{ 
  int j, valid;
  double *weights, T, orig_T, temp;
  thnode_t *th;
  node_t *fanin;
  input_phase_t phase;
  FILE *fp;    
    
  /* initially set the weight-threshold solution to be invalid */
  /* finally, if all weights and threshold are 0, valid = 0 */
  valid = 0;
  /* read in the threshold value */
  if ((fp = fopen("thsyn_ampl.out", "r")) == NULL)
    fail("thsyn_create_thnode: unable to open file handle for writing");

  fscanf(fp, "%lf ", &T);
  orig_T = T;/* save the original value of T */

  /* check if this solution is valid, if this lp problem has no solution,the returned values are all zeros */
  if (T != 0)
    valid = 1;

  /* read from file and assign the values of weights and threshold */
  if (combine_case == THSYN_NO_COMBINE) 
    weights = ALLOC(double, node_num_fanin(n));
  else /* it is THSYN_OR_COMBINE case */
    weights = ALLOC(double, node_num_fanin(n) + 1);
  
  foreach_fanin(n, j, fanin) {
    phase = node_input_phase(n, fanin);
    
    fscanf(fp,"%lf ",&temp);
    if (phase == POS_UNATE)
      weights[j] = temp; 
    else if (phase == NEG_UNATE) {
      weights[j] = -temp; 
      T = T + weights[j]; 
    }
    else
      fail("thsyn_create_thnode: binate phase in a unate node");
    
    if ((valid == 0) && weights[j] != 0)
      valid = 1;
  }

  fclose(fp);

  /* create thnode */
  th = thnode_alloc();

  /* check if it is a valid threshold node */
  if (valid == 1) { /* a valid threshold node */
    if (combine_case == THSYN_NO_COMBINE) { /* i.e., the node itself is a threshold node */
      th->nin = node_num_fanin(n);
      th->W = weights;
      th->T = T;
    }
    else { /* i.e., combination is needed to create a larger threshold node */
      if (combine_case == THSYN_OR_COMBINE) {
	th->nin = node_num_fanin(n) + 1;
	weights[node_num_fanin(n)] = orig_T + (double)(opt->delta_on);
	th->W = weights;
	th->T = T;
      }
      else
	fail("thsyn_create_thnode: meet combine case other than THSYN_NO_COMBINE and THSYN_OR_COMBINE.");
    }
  }
  else { /* a invalid threshold node */
    thnode_free(th);
    th = NULL;
  }
  return th;
}

/**************************************************************
 * since n is not a threshold node, split into two nodes and  *
 * recursively check whether the splitted nodes are threshold *
 **************************************************************/
void thsyn_unate_split(network_t *network, thsyn_options_t *opt, node_t *node)
{
  int  case_index, i, num_fanin;
  node_split_t *split;
  node_t *sn1, *sn2, *sn1_literal, *combined_node, *tmp_node;
  thnode_t *th;

  /* check if this node exists in the current threshold network or not */
  if (!st_lookup(threshold_nodes, node_name(node), NIL(char *))) {
    /* split node */
    split = thsyn_split_unate_node(node, 1 /*allow and-split case*/);
    /* set the index: whether it was an OR split or AND split */
    if (split->split_flag)
      case_index = THSYN_AND_COMBINE;
    else
      case_index = THSYN_OR_COMBINE;
    
    /* sort the nodes and get the splitted nodes in descending order */
    thsyn_sort_nodes(split->nodes);
    sn1 = array_fetch(node_t *, split->nodes, 0);
    sn2 = array_fetch(node_t *, split->nodes, 1);
    
    /* check if sn1 or sn2 is unate, fail if we get BINATE */
    if ((thsyn_check_unate(sn1) == BINATE) || (thsyn_check_unate(sn2) == BINATE))
      fail("thsyn_unate_split: encountered binate node in unate split");
    
    /* check if sn2 fans out to only one primary input, if it is, for example, 
       f = sn1 sn2, where sn2 fans out to only one primary input: sn2 = pa,
       store f: sn1, pa, [2 1 1], then process fanins of sn1 recursively. 
       This can't happen for sn1, since sn1 and sn2 are sorted in descending order,
       If sn1 only has one fanin, sn2 can only has one fanin, and all two fanin nodes
       are threshold node. */
    if (node_num_fanin(sn2) == 1) {
      sn1_literal = node_literal(sn1, 1); /* 0: complemented, 1: uncomplemented */
      
      if (case_index == THSYN_OR_COMBINE)  /* OR case */
	combined_node = node_or(sn1_literal, sn2);
      else /* AND case */
	combined_node = node_and(sn1_literal, sn2);
      
      FREE(combined_node->name);
      combined_node->name = util_strsav(split->pn_name);
            
      /* write AMPL data file  */
      thsyn_write_ampl_data(combined_node, opt);
      
      /* system call to AMPL */
      if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	fail("thsyn_unate_split: system call failed.");
      
      th = thsyn_create_thnode(combined_node, opt, THSYN_NO_COMBINE); 
  
      if (th == NIL(thnode_t)) {
      	fail("thsyn_unate_split: AMPL failed to solve a 2-input OR gate");
      }
      else {
	/* create a threshold node based on thnode, and add this thnode into threshold network */
	tmp_node = node_dup(combined_node);
	/* attach the thnode to the node */
	tmp_node->th = (char *) th;
	FREE(tmp_node->name);
	tmp_node->name = util_strsav(node_name(combined_node));
	/* check if this node exists in the current threshold network or not */
	if (!st_lookup(threshold_nodes, node_name(tmp_node), NIL(char *))) {
	  /* save this threshold node in threshold network */
	  st_insert(threshold_nodes, node_name(tmp_node), (char *) tmp_node);
	}
      }
    }
    else {
      /* for sn1 and sn2 are nodes that have more than one fanin, call AMPL on the larger node,
         check if it is a threshold node, if not, call AMPL on the smaller node, and check again */
      if (case_index != THSYN_OR_COMBINE)
	fail("meet THSYN_AND_COMBINE case, which is impossible here");
      
      /* write AMPL data file for sn1 */
      thsyn_write_ampl_data(sn1, opt);
      /* system call to AMPL */
      if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	fail("thsyn_unate_split: system call failed.");
      
      th = thsyn_create_thnode(sn1, opt, THSYN_OR_COMBINE); /* it is always THSYN_OR_COMBINE case */
      if (th != NIL(thnode_t))
	/* sn1 is a threshold node, then get new threshold pn_name = sn1(...) + sn2 */
	combined_node = thsyn_unate_combine(sn1, sn2, th, split, opt);
      else { /* sn1 is not a threshold node, go to check sn2 */
	/* write AMPL data file for sn2 */
	thsyn_write_ampl_data(sn2, opt);
	/* system call to AMPL */
	if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
	  fail("thsyn_unate_split: system call failed.");
	
	th = thsyn_create_thnode(sn2, opt, THSYN_OR_COMBINE); /* it is always THSYN_OR_COMBINE case */
	if (th != NIL(thnode_t))
	  /* sn2 is a threshold node, then get new threshold pn_name = sn1 + sn2(...) */
	  combined_node = thsyn_unate_combine(sn2, sn1, th, split, opt);
	else { /* both sn1 and sn2 are not threshold case */
	  combined_node = thsyn_split_w_fanin(node, opt);
	}

      }
    }
    /* for each fanin of tmp_node, start thsyn_one_node again */
    num_fanin = node_num_fanin(combined_node);
    for ( i = 0; i < num_fanin; i++)
      thsyn_one_node(network, opt, node_get_fanin(combined_node, i));
  }
}

node_t *new_thsyn_unate_combine(node_t *node_function, node_t *literal, thnode_t *th, node_split_t *split, thsyn_options_t *opt)
{
  node_t *combined_node;

  literal = node_literal(literal, 1); /* 0: complemented, 1: uncomplemented */
  combined_node = node_or(node_function, literal); /* OR_COMBINE case */
  FREE(combined_node->name);
  combined_node->name = util_strsav(split->pn_name);
  /* attach the thnode to the node */
  combined_node->th = (char *) th;

  return combined_node;
}


node_t *thsyn_unate_combine(node_t *node_function, node_t *literal, thnode_t *th, node_split_t *split, thsyn_options_t *opt)
{
  node_t *tmp_node, *combined_node;

  literal = node_literal(literal, 1); /* 0: complemented, 1: uncomplemented */
  combined_node = node_or(node_function, literal); /* OR_COMBINE case */
  FREE(combined_node->name);
  combined_node->name = util_strsav(split->pn_name);
  /* create a threshold node based on thnode, and add this thnode into threshold network */
  tmp_node = node_dup(combined_node);
  /* attach the thnode to the node */
  tmp_node->th = (char *) th;
  FREE(tmp_node->name);
  tmp_node->name = util_strsav(node_name(combined_node));
  /* check if this node exists in the current threshold network or not */
  if (!st_lookup(threshold_nodes, node_name(tmp_node), NIL(char *))) {
    /* save this threshold node in threshold network */
    st_insert(threshold_nodes, node_name(tmp_node), (char *) tmp_node);
  }
  return combined_node;
}

node_t *new_thsyn_split_w_fanin(node_t *node, thsyn_options_t *opt)
{
  int i, is_unate, num_cubes, max, num_nodes;
  char *pn_name;
  lsList l1, l2;
  lsGen g1, g2;
  array_t *tmp_vec, *array, *array_copy;
  node_t *p, *t, *combined_node, *tmp_node;
  node_split_t *split;
  thnode_t *th;


  pn_name = util_strsav(node_name(node));
  num_cubes = node_num_cube(node);
  max = num_cubes;

  l1 = lsCreate();
  l2 = lsCreate();

  lsNewBegin(l1, (lsGeneric) node, 0);
  num_nodes = lsLength(l1);

  tmp_vec = array_alloc(node_t *, 2);

  while ((num_nodes <= (max - 1)) && lsDelBegin(l1, (lsGeneric *) &p) != LS_NOMORE) {
    if (node_num_cube(p) == 1)
      lsNewEnd(l2, (lsGeneric) p, 0);
    else {
      is_unate = thsyn_check_unate(p);
      
      if (is_unate == BINATE)
	tmp_vec = thsyn_split_binate_node(p);
      else {
	split = thsyn_split_unate_node(p, 0 /* do not allow and-split case */);
	tmp_vec = split->nodes; 
      }
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 0), 0);
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 1), 0);
    }
    num_nodes = lsLength(l1) + lsLength(l2);
  }

  g1 = lsStart(l1);
  while (lsNext(g1, (lsGeneric *) &p, 0) != LS_NOMORE)
    lsNewEnd(l2, (lsGeneric) p, 0);
  lsFinish(g1);

  array = array_alloc(node_t *, lsLength(l2));

  g2 = lsStart(l2);
  while (lsNext(g2, (lsGeneric *) &p, 0) != LS_NOMORE)
    array_insert_last(node_t *, array, p);
  
  lsFinish(g2);
  array_copy = array_alloc(node_t *, 0);
  for (i = 0; i < array_n(array); i++) {
    t = array_fetch(node_t *, array, i);
    if ((node_num_fanin(t) == 1) && (node_input_phase(t, node_get_fanin(t, 0)) == POS_UNATE))
      array_insert_last(node_t *, array_copy, node_get_fanin(t, 0));
    else
      array_insert_last(node_t *, array_copy, t);
  }
  
  array_free(tmp_vec);
  array_free(array);
  
  /* get all the literals of the split nodes, and OR them together to get the m-input OR gate, 
     and also create the corresponding threshold node for it and save the threshold node into 
     threshold network */
  combined_node = node_literal(array_fetch(node_t *, array_copy, 0), 1); /* 0: complemented, 1: uncomplemented */
  for (i = 1; i < array_n(array_copy); i++)
    combined_node = node_or(combined_node, node_literal(array_fetch(node_t *, array_copy, i), 1));
  
  /* write AMPL data file for combined_node */
  thsyn_write_ampl_data(combined_node, opt);
  /* system call to AMPL */

  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
    fail("new_thsyn_split_w_fanin: system call failed.");

  th = thsyn_create_thnode(combined_node, opt, THSYN_NO_COMBINE); 
  if (th != NIL(thnode_t)) { 
    tmp_node = node_dup(combined_node);
    FREE(tmp_node->name);
    tmp_node->name = util_strsav(pn_name);
    /* attach the thnode to the node */
    tmp_node->th = (char *) th;
  }
  else
    fail("in new_thsyn_split_w_fanin, AMPL retuned no solution for an m-input OR gate");
  
  return tmp_node; 
}

/*****************************************************************************
 * This function split a node (unate or binate) into m-input OR gate, where  *
 * m = (num_cubes <= opt->max_fanin) ? num_cubes:opt->max_fanin. The m-input *
 * OR gate is returned, and the m split nodes are saved in array_copy        *
 *****************************************************************************/
node_t *thsyn_split_w_fanin(node_t *node, thsyn_options_t *opt)
{
  int i, is_unate, num_cubes, max, num_nodes;
  char *pn_name;
  lsList l1, l2;
  lsGen g1, g2;
  array_t *tmp_vec, *array, *array_copy;
  node_t *p, *t, *combined_node, *tmp_node;
  node_split_t *split;
  thnode_t *th;


  pn_name = util_strsav(node_name(node));
  num_cubes = node_num_cube(node);
  max = (num_cubes <= opt->max_fanin) ? num_cubes : opt->max_fanin;

  l1 = lsCreate();
  l2 = lsCreate();

  lsNewBegin(l1, (lsGeneric) node, 0);
  num_nodes = lsLength(l1);

  tmp_vec = array_alloc(node_t *, 2);

  while ((num_nodes <= (max - 1)) && lsDelBegin(l1, (lsGeneric *) &p) != LS_NOMORE) {
    if (node_num_cube(p) == 1)
      lsNewEnd(l2, (lsGeneric) p, 0);
    else {
      is_unate = thsyn_check_unate(p);
      
      if (is_unate == BINATE)
	tmp_vec = thsyn_split_binate_node(p);
      else {
	split = thsyn_split_unate_node(p, 0 /* do not allow and-split case */);
	tmp_vec = split->nodes; 
      }
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 0), 0);
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 1), 0);
    }
    num_nodes = lsLength(l1) + lsLength(l2);
  }

  g1 = lsStart(l1);
  while (lsNext(g1, (lsGeneric *) &p, 0) != LS_NOMORE)
    lsNewEnd(l2, (lsGeneric) p, 0);
  lsFinish(g1);

  array = array_alloc(node_t *, lsLength(l2));

  g2 = lsStart(l2);
  while (lsNext(g2, (lsGeneric *) &p, 0) != LS_NOMORE)
    array_insert_last(node_t *, array, p);
  
  lsFinish(g2);
  array_copy = array_alloc(node_t *, 0);
  for (i = 0; i < array_n(array); i++) {
    t = array_fetch(node_t *, array, i);
    if ((node_num_fanin(t) == 1) && (node_input_phase(t, node_get_fanin(t, 0)) == POS_UNATE))
      array_insert_last(node_t *, array_copy, node_get_fanin(t, 0));
    else
      array_insert_last(node_t *, array_copy, t);
  }
  
  array_free(tmp_vec);
  array_free(array);
  
  /* get all the literals of the split nodes, and OR them together to get the m-input OR gate, 
     and also create the corresponding threshold node for it and save the threshold node into 
     threshold network */
  combined_node = node_literal(array_fetch(node_t *, array_copy, 0), 1); /* 0: complemented, 1: uncomplemented */
  for (i = 1; i < array_n(array_copy); i++)
    combined_node = node_or(combined_node, node_literal(array_fetch(node_t *, array_copy, i), 1));
  
  /* write AMPL data file for combined_node */
  thsyn_write_ampl_data(combined_node, opt);
  /* system call to AMPL */

  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
    fail("thsyn_split_w_fanin: system call failed.");

  th = thsyn_create_thnode(combined_node, opt, THSYN_NO_COMBINE); 
  if (th != NIL(thnode_t)) { 
    tmp_node = node_dup(combined_node);
    FREE(tmp_node->name);
    tmp_node->name = util_strsav(pn_name);
  
    /* attach the thnode to the node */
    tmp_node->th = (char *) th;
    /* check if this node exists in the current threshold network or not */
    if (!st_lookup(threshold_nodes, node_name(tmp_node), NIL(char *))) {
    /* save this threshold node in threshold network */
      st_insert(threshold_nodes, node_name(tmp_node), (char *) tmp_node);
    }
  }
  else
    fail("in thsyn_split_w_fanin, AMPL retuned no solution for an m-input OR gate");
  
  return combined_node; 
}

/************************************************************************************** 
 *  This function splits a binate node into two parts.                                *
 *  The splitting is based on the binate variables that appears most frequently.      *
 *  And the splitted nodes are returned in an array.                                  *
 *  For example:                                                                      *
 *  binate function: f = a b + a c + a' b' + a' d + e,  where binate variable appears *
 *  most frequently. f can be written as: f = a f1 + a' f2 + f3, where f1 = b + c,    *
 *  f2 = b' + d, f3 = e, after splitting we get two nodes: f = sn1 + sn1,             *
 *  where sn1 = a f1 = a b + a c, sn2 = a' f2 + f3 = a' b' + a' d + e                 *
 **************************************************************************************/
array_t *thsyn_split_binate_node(node_t *node)
{
  int i, num_fanin, max_index, max_freq, fanin_index, *literals;
  node_t *fanin, *f1, *f2, *f3, *s0, *s1, *p1, *p2, *p23;
  array_t *split;

  /* get number of fanin */
  num_fanin = node_num_fanin(node);

  /* sanity check */
  if (num_fanin <= 0)
    fail("fanin less than/equal to zero for binate node");

  /* get the literal count of each variable in node */
  literals = node_literal_count(node);

  /* initialize the max index to -1 and max freq to 0 */
  max_index = fanin_index = -1;
  max_freq = 0;

  for (i = 0; i < 2*num_fanin; i++) {
    /* if even, then increment fanin index */
    if (!(i%2))
      fanin_index++;

    /* get fanin and particular index and figure out what phase it appears in within node */
    fanin = node_get_fanin(node, fanin_index);
    switch (node_input_phase(node, fanin)) {
    case POS_UNATE: break;
    case NEG_UNATE: break;
    case BINATE: // binate
      /* if this literal count is bigger than what we have seen so far, update the 
	 count and the fanin_index (i.e. the variable which leads to this count) */
      if (literals[i] >= max_freq) {
	max_freq = literals[i];
	max_index = fanin_index;
      }
      break;
    default:
      fail("binate split phase unknown");
    }
  }
  
  /* check to see that we have a valid index */
  if (max_index < 0)
    fail("max index is less than zero in binate split");

  /* we split according to this binate node */
  fanin = node_get_fanin(node, max_index);
  node_algebraic_cofactor(node, fanin, &f1, &f2, &f3);
  
  s0 = node_literal(fanin, 0);
  s1 = node_literal(fanin, 1);

  /* a f1 */
  p1 = node_and(f1, s1);
  /* a' f2 */
  p2 = node_and(f2, s0);
  /* a' f2 + f3 */
  p23 = node_or(p2, f3);

  FREE(s0);
  FREE(s1);
  FREE(p2);
  FREE(f1);
  FREE(f2);
  FREE(f3);
  FREE(literals);

  p1->type = INTERNAL;
  p23->type = INTERNAL;

  split = array_alloc(node_t *, 2);
  array_insert(node_t *, split, 0, p1);
  array_insert(node_t *, split, 1, p23);
  return split;
}

/*************************************************************************************
 * Given an unate node as a parameter to this function,it splits this node which is  *
 * not threshold gate into two nodes based on the most appearing literal.            *  
 * Three cases of splitting happens here:                                            *
 * Case 1:                                                                           *
 * all variables appears only once, then, split this node by half                    *
 * and these two splitted nodes are ORed together                                    *
 * e.g   f = a b + c d + e f,                                                        *
 * after splitting: f = sn1 + sn2 where sn1 = a b, sn2 = c d + e f                   *
 * Case 2:                                                                           *
 * There is at least of one variable that appears in each product terms              *
 * of the logic function, then split this variable in one node, and                  *
 * the left are in the other node. These two splitted nodes are ANDed together.      *  
 * e.g   f = a b + a c + a d,                                                        *
 * after splitting: f = sn1 sn2 where sn1 = a, sn2 = b + c + d                       *
 * case 3:                                                                           *
 * if it is not case 1 and case 2, split the node into two nodes as following:       *
 * all prime implicants which contain the most frequently appeared variable for the  *
 * first part and the others form the second part. These two nodes are ORed together.*
 * e.g  f = a b + a c + d e + f,                                                     *
 * after splitting: f = sn1 + sn2, where sn1 = a b + a c, sn2 = d e + f              *
 *************************************************************************************/
node_split_t *thsyn_split_unate_node(node_t *node, int and_allow)
{
  int num_fanin, max_index, max_freq, fanin_index, i, j, k, min_index, min_freq;
  int *literals;
  node_split_t *result;
  node_t *fanin, *f0, *f1, *p1, *p2, *tmp0, *tmp1;
  node_cube_t c;
  node_literal_t l;


  num_fanin = node_num_fanin(node); /* get number of fanin */

  if (num_fanin <= 0) /* sanity check */
    fail("thsyn_split_unate_node: fanin less than/equal to zero for unate node");

  /* initialize */
  result = ALLOC(node_split_t, 1);
  result->split_flag = 0;
  result->pn_name = util_strsav(node_name(node));
  /* get the literal count of each variable in node */
  literals = node_literal_count(node);
  /* initialize the max index to 0 (the first literal) and max freq to the contents 
     of the first literal */
  max_index = fanin_index = 0;
  max_freq = literals[0];
  /* find most frequently appearing literal in all fanins */
  for (i = 1; i < 2 * num_fanin; i++) {
    if (!(i%2))
      fanin_index++;

    if (literals[i] > max_freq) {
      max_freq = literals[i];
      max_index = fanin_index;
    }
  }

  /* initialize the min index to -1 and min freq to max_freq + 1
     note here!, the min_freq can be 0 for some fanins, but we dont 
     consider such fanins*/
  min_index = fanin_index = -1;
  min_freq = max_freq + 1;
  /* find least frequently appearing literal in all fanins */
  for (i = 0; i < 2 * num_fanin; i++) {
    if (!(i%2))
      fanin_index++;
    if (literals[i] != 0) {
      if (literals[i] < min_freq) {
	min_freq = literals[i];
	min_index = fanin_index;
      }
    }
  }

  FREE(literals);
  
  if (max_freq > 1) {
    /* compute p2 */
    fanin = node_get_fanin(node, max_index);
    f0 = node_literal(fanin, 0); /* complemented */
    f1 = node_literal(fanin, 1); /* uncomplemented */
    
    switch(node_input_phase(node, fanin)) {
    case POS_UNATE:
      p2 = node_cofactor(node, f0);
      break;
    case NEG_UNATE:
      p2 = node_cofactor(node, f1);
      break;
    default:
      fail("not a unate function");
    }
    
    FREE(f0);
    FREE(f1);
  }
  else { /* max_freq is 1 */
    /* when each variable appears at most once */
    k = node_num_cube(node);
    /* compute p2 */
    p2 = node;
    for (i = 0; i < k; i++) {
      c = node_get_cube(node, i);
      foreach_fanin(node, j, fanin) {
	l = node_get_literal(c, j);
	if (l == ONE) {
	  f0 = node_literal(fanin, 0); /* uncomplemented */
	  p2 = node_cofactor(p2, f0);
	  FREE(f0);
	}
	else if (l == ZERO) {
	  f1 = node_literal(fanin, 1); /* complemented */
	  p2 = node_cofactor(p2, f1);
	  FREE(f1);
	}
	else if (l == TWO) {
	  continue;
	}
	else
	  fail("bad literal");
	
	if ((i+1) >= (int) k/2)
	  break;
      }
      if ((i+1) >= (int) k/2)
	break;
    }
  }
  

  /* special case: when all prime implicants contain the most
     prime implicants e.g. x1x2 + x2x3 + x2x4 + x2x5 */
  /* if and_allow == 1, then split like this to get: p1 = x2 and 
     p2 = x1 + x3 + x4 + x5 */
  /* if and_allow == 0, then split into two: p1 = x1 x2 and 
     p2 = x2x3 + x2x4 + x2x5 */

  if ((max_freq > 1) && (node_num_fanin(p2) == 0)) {
    FREE(p2);

    if (and_allow == 1) {
      /* recompute p1 and p2 */
      /* first set split_flag so that the result is and_combined */
      result->split_flag = 1;
      fanin = node_get_fanin(node, max_index);
      switch(node_input_phase(node, fanin)) {
      case POS_UNATE:
	p1 = node_literal(fanin, 1); /* uncomplemented */
	break;
      case NEG_UNATE:
	p1 = node_literal(fanin, 0); /* complemented */
	break;
      default:
	fail("not a unate function");
      }
      p2 = node_cofactor(node, p1);
    }
    else { /* only or_split method is allowed */
      /* recompute p1 and p2 */
      result->split_flag = 0;
      fanin = node_get_fanin(node, min_index);
      f0 = node_literal(fanin, 0); /* complemented */
      f1 = node_literal(fanin, 1); /* uncomplemented */
      
      switch(node_input_phase(node, fanin)) {
      case POS_UNATE:
	p2 = node_cofactor(node, f0);
	break;
      case NEG_UNATE:
	p2 = node_cofactor(node, f1);
	break;
      default:
	fail("not a unate function");
      }
      
      FREE(f0);
      FREE(f1);
      /* compute p1 */
      p1 = node_dup(node);
      tmp0 = node_dup(p2);
      node_substitute(p1, tmp0, 0);
      f0 = node_literal(tmp0, 0);
      tmp1 = node_cofactor(p1, f0);
      FREE(p1);
      FREE(tmp0);
      FREE(f0);
      p1 = tmp1;
    }
  }
  else {
    /* compute p1 */
    p1 = node_dup(node);
    tmp0 = node_dup(p2);
    node_substitute(p1, tmp0, 0);
    f0 = node_literal(tmp0, 0);
    tmp1 = node_cofactor(p1, f0);
    FREE(p1);
    FREE(tmp0);
    FREE(f0);
    p1 = tmp1;
  }
  
  if ((node_num_fanin(p1) == 0) && (node_type(p1) != PRIMARY_INPUT))
    fail("p1 is empty");
  if ((node_num_fanin(p2) == 0) && (node_type(p2) != PRIMARY_INPUT))
    fail("p2 is empty");

  p1->type = INTERNAL;
  p2->type = INTERNAL;
  
  result->nodes = array_alloc(node_t *, 2);
  array_insert(node_t *, result->nodes, 0, p1);
  array_insert(node_t *, result->nodes, 1, p2);
  
  return result;
}


node_t *new_thsyn_binate_split_m2(node_t *node, thsyn_options_t *opt)
{
  int i, is_unate, num_cubes, max, num_nodes;
  char *pn_name;
  lsList l1, l2;
  lsGen g1, g2;
  array_t *tmp_vec, *array, *array_copy;
  node_t *p, *t, *combined_node, *tmp_node;
  thnode_t *th;

  pn_name = util_strsav(node_name(node));
  num_cubes = node_num_cube(node);
  max = (num_cubes <= opt->max_fanin) ? num_cubes : opt->max_fanin;

  l1 = lsCreate();
  l2 = lsCreate();
  lsNewBegin(l1, (lsGeneric) node, 0);
  num_nodes = lsLength(l1);

  tmp_vec = array_alloc(node_t *, 2);

  while ((num_nodes <= (max - 1)) && lsDelBegin(l1, (lsGeneric *) &p) != LS_NOMORE) {
    if (node_num_cube(p) == 1) {
      lsNewEnd(l2, (lsGeneric) p, 0);
      break;
    }
    else {
      is_unate = thsyn_check_unate(p);
      
      if (is_unate == BINATE)
	tmp_vec = thsyn_split_binate_node(p);
      else {
	lsNewEnd(l1, (lsGeneric) p, 0);
	break;
      }
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 0), 0);
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 1), 0);
    }
    num_nodes = lsLength(l1) + lsLength(l2);
  }

  g1 = lsStart(l1);
  while (lsNext(g1, (lsGeneric *) &p, 0) != LS_NOMORE)
    lsNewEnd(l2, (lsGeneric) p, 0);
  lsFinish(g1);

  array = array_alloc(node_t *, 0);
  g2 = lsStart(l2);
  while (lsNext(g2, (lsGeneric *) &p, 0) != LS_NOMORE)
    array_insert_last(node_t *, array, p);
  lsFinish(g2);

  array_copy = array_alloc(node_t *, 0);
  for (i = 0; i < array_n(array); i++) {
    t = array_fetch(node_t *, array, i);
    if ((node_num_fanin(t) == 1) && (node_input_phase(t, node_get_fanin(t, 0)) == POS_UNATE))
      array_insert_last(node_t *, array_copy, node_get_fanin(t,0));
    else
      array_insert_last(node_t *, array_copy, t);
  }
  
  array_free(tmp_vec);
  array_free(array);

  /* get all the literals of the split nodes, and OR them together to get the m-input OR gate, 
     and also create the corresponding threshold node for it and return this combined node */
  combined_node = node_literal(array_fetch(node_t *, array_copy, 0), 1); /* 0: complemented, 1: uncomplemented */
  for (i = 1; i < array_n(array_copy); i++) {
    combined_node = node_or(combined_node,node_literal(array_fetch(node_t *, array_copy, i), 1));
  }

  /* write AMPL data file for combined_node */
  thsyn_write_ampl_data(combined_node, opt);
  /* system call to AMPL */
  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
    fail("new_thsyn_binate_split_m2: system call failed.");

  th = thsyn_create_thnode(combined_node, opt, THSYN_NO_COMBINE); 
  if (th != NIL(thnode_t)) {
    tmp_node = node_dup(combined_node);
    FREE(tmp_node->name);
    tmp_node->name = util_strsav(pn_name);
    /* attach the thnode to the node */
    tmp_node->th = (char *) th;
  }
  else {
    fail("in new_thsyn_binate_split_m2, AMPL retuned no solution for an m-input OR gate");
  }


  return tmp_node; 
}

/******************************************************************************
 * This function split a node (unate or binate) into k-input OR gate, k <=m,  *
 * where m = (num_cubes <= opt->max_fanin) ? num_cubes:opt->max_fanin.        *
 * Splitting stops when it reaches a unate node (called method 2 (m2)). The   *
 * k-input OR gate is returned, and the k split nodes are saved in array_copy *
 ******************************************************************************/
node_t *thsyn_binate_split_m2(node_t *node, thsyn_options_t *opt)
{
  int i, is_unate, num_cubes, max, num_nodes;
  char *pn_name;
  lsList l1, l2;
  lsGen g1, g2;
  array_t *tmp_vec, *array, *array_copy;
  node_t *p, *t, *combined_node, *tmp_node;
  thnode_t *th;

  pn_name = util_strsav(node_name(node));
  num_cubes = node_num_cube(node);
  max = (num_cubes <= opt->max_fanin) ? num_cubes : opt->max_fanin;

  l1 = lsCreate();
  l2 = lsCreate();
  lsNewBegin(l1, (lsGeneric) node, 0);
  num_nodes = lsLength(l1);

  tmp_vec = array_alloc(node_t *, 2);

  while ((num_nodes <= (max - 1)) && lsDelBegin(l1, (lsGeneric *) &p) != LS_NOMORE) {
    if (node_num_cube(p) == 1) {
      lsNewEnd(l2, (lsGeneric) p, 0);
      break;
    }
    else {
      is_unate = thsyn_check_unate(p);
      
      if (is_unate == BINATE)
	tmp_vec = thsyn_split_binate_node(p);
      else {
	lsNewEnd(l1, (lsGeneric) p, 0);
	break;
      }
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 0), 0);
      lsNewEnd(l1, (lsGeneric) array_fetch(node_t *, tmp_vec, 1), 0);
    }
    num_nodes = lsLength(l1) + lsLength(l2);
  }

  g1 = lsStart(l1);
  while (lsNext(g1, (lsGeneric *) &p, 0) != LS_NOMORE)
    lsNewEnd(l2, (lsGeneric) p, 0);
  lsFinish(g1);

  array = array_alloc(node_t *, 0);
  g2 = lsStart(l2);
  while (lsNext(g2, (lsGeneric *) &p, 0) != LS_NOMORE)
    array_insert_last(node_t *, array, p);
  lsFinish(g2);

  array_copy = array_alloc(node_t *, 0);
  for (i = 0; i < array_n(array); i++) {
    t = array_fetch(node_t *, array, i);
    if ((node_num_fanin(t) == 1) && (node_input_phase(t, node_get_fanin(t, 0)) == POS_UNATE))
      array_insert_last(node_t *, array_copy, node_get_fanin(t,0));
    else
      array_insert_last(node_t *, array_copy, t);
  }
  
  array_free(tmp_vec);
  array_free(array);

  /* get all the literals of the split nodes, and OR them together to get the m-input OR gate, 
     and also create the corresponding threshold node for it and save the threshold node into 
     threshold network */
  combined_node = node_literal(array_fetch(node_t *, array_copy, 0), 1); /* 0: complemented, 1: uncomplemented */
  for (i = 1; i < array_n(array_copy); i++) {
    combined_node = node_or(combined_node,node_literal(array_fetch(node_t *, array_copy, i), 1));
  }

  /* write AMPL data file for combined_node */
  thsyn_write_ampl_data(combined_node, opt);
  /* system call to AMPL */
  if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
    fail("thsyn_binate_split_m2: system call failed.");

  th = thsyn_create_thnode(combined_node, opt, THSYN_NO_COMBINE); 
  if (th != NIL(thnode_t)) {
    tmp_node = node_dup(combined_node);
    FREE(tmp_node->name);
    tmp_node->name = util_strsav(pn_name);
  
    /* attach the thnode to the node */
    tmp_node->th = (char *) th;
    /* check if this node exists in the current threshold network or not */
    if (!st_lookup(threshold_nodes, node_name(tmp_node), NIL(char *))) {
      /* save this threshold node in threshold network */
      st_insert(threshold_nodes, node_name(tmp_node), (char *) tmp_node);
    }
  }
  else {
    fail("in thsyn_binate_split_m2, AMPL retuned no solution for an m-input OR gate");
  }


  return combined_node; 
}

/*******************************************************************
 * fixes up the fanin pointers of a node to those of the new nodes *
 * that have been created during synthesis.                        * 
 *******************************************************************/
static void new_fixup_fanin(void)
{
  array_t *thnodes;
  int i, j;
  char *key;
  node_t *n, *v, *fanin;
  st_generator *gen;

  thnodes = array_alloc(node_t *, 0);
  st_foreach_item(new_threshold_nodes, gen, &key, (char **) &n)
    array_insert_last(node_t *, thnodes, n);

  for (i = 0; i < array_n(thnodes); i++) {
    n = array_fetch(node_t *, thnodes, i);
    foreach_fanin(n, j, fanin) {
      if (st_lookup(new_threshold_nodes, node_name(fanin), (char **) &v))
	n->fanin[j] = v;
    }
  }
  
  array_free(thnodes);
}


/*******************************************************************
 * fixes up the fanin pointers of a node to those of the new nodes *
 * that have been created during synthesis.                        * 
 *******************************************************************/
static void fixup_fanin(void)
{
  array_t *thnodes;
  int i, j;
  char *key;
  node_t *n, *v, *fanin;
  st_generator *gen;

  thnodes = array_alloc(node_t *, 0);
  st_foreach_item(threshold_nodes, gen, &key, (char **) &n)
    array_insert_last(node_t *, thnodes, n);

  for (i = 0; i < array_n(thnodes); i++) {
    n = array_fetch(node_t *, thnodes, i);
    foreach_fanin(n, j, fanin) {
      if (st_lookup(threshold_nodes, node_name(fanin), (char **) &v))
	n->fanin[j] = v;
    }
  }
  
  array_free(thnodes);
}

/***************
 * thsyn usage *
 ***************/
static int thsyn_usage(void)
{
  fprintf(siserr, "usage: thsyn [-c num1] [-n num2] [-f num3] [-o file] [-h]\n"); 
  fprintf(siserr, "   -c num1:\tAssign the integer num1 to fanin restriction, default value is 6.\n");
  fprintf(siserr, "   -n num2:\tAssign the integer num2 to delta_on, default value is 0.\n");
  fprintf(siserr, "   -f num3:\tAssign the integer num3 to delta_off, default value is 1.\n");
  fprintf(siserr, "   -o file:   \tWrite the linear program problem into file.mod and file.dat,\n");
  fprintf(siserr, "              \tdefault write to thsyn_ampl.mod and thsyn_ampl.dat.\n");
  fprintf(siserr, "   -h:        \tFor help information. \n");  
  return TH_FAIL;
}

static int thsyn_red_remove_usage(void)
{
  fprintf(siserr, "usage: thsyn_red_remove [-c num1] [-n num2] [-f num3] [-o file] [-h]\n"); 
  fprintf(siserr, "   -c num1:\tAssign the integer num1 to fanin restriction, default value is 6.\n");
  fprintf(siserr, "   -n num2:\tAssign the integer num2 to delta_on, default value is 0.\n");
  fprintf(siserr, "   -f num3:\tAssign the integer num3 to delta_off, default value is 1.\n");
  fprintf(siserr, "   -o file:   \tWrite the linear program problem into file.mod and file.dat,\n");
  fprintf(siserr, "              \tdefault write to thsyn_ampl.mod and thsyn_ampl.dat.\n");
  fprintf(siserr, "   -h:        \tFor help information. \n");  
  return TH_FAIL;
}

